// kbukint9x.cpp : United Kingdom International Keyboard Layout for 9x OSs
//   (c) 2004 John Sullivan
//
// Generate and register a Win9x .KBD file
//

#include "stdafx.h"
#include "kbgen.h"

//
// Definitions
//

const TCHAR kbdFile[] = _T("KBUKINT");
const TCHAR kbdDesc[] = _T("United Kingdom (International)");
const unsigned short kbdLCID = 0x0809;


//
// helper class
//
// The .KBD file format is a bit messy, with lots of internal pointers,
// offsets and lengths. These potentially all change whenever the layout
// is tweaked. To avoid having to manage all this by hand we build a nested
// array of byte-streams, which can track its own internal structure and cause
// the correct values to magically appear in the middle of the data file.
//

#include <vector>

class ByteArray
{
public:
    ByteArray()
    {
        m_buf = (BYTE*)malloc(1024);
        memset(m_buf, 0, 1024);
        m_buflen = 1024;
        m_len = 0;
        m_fulllen = 0;
    }

    ~ByteArray()
    {
        free(m_buf);
    }

    WORD Length(){ return m_len; }
    WORD FullLength(){ return m_fulllen; }
    operator const BYTE*(){ return m_buf; }

    void outb(BYTE v)
    {
        if (m_len>=m_buflen) {
            m_buf = (BYTE*)realloc(m_buf, m_buflen+1024);
            memset(m_buf+m_buflen, 0, 1024);
            m_buflen += 1024;
        }

        m_buf[m_len++] = v;
    }

    void outmb(BYTE v, int n)
    {
        for ( int i=0 ; i<n ; i++ ) {
            outb(v);
        }
    }

    void outstr(const char *v)
    {
        for ( int i=0 ; v[i] ; i++ ) {
            outb(v[i]);
        }
    }

    void outstrn(const char *v, int n)
    {
        for ( int i=0 ; i<n ; i++ ) {
            outb(v[i]);
        }
    }

    void outw(WORD v)
    {
        outb(v & 0xff);
        outb((v>>8) & 0xff);
    }

    void outdw(DWORD v)
    {
        outw((WORD)(v & 0xffff));
        outw((WORD)((v>>16) & 0xffff));
    }

    void outar(ByteArray *ar)
    {
        m_subars.push_back(ar);
    }

    void outaroffb(ByteArray *org, ByteArray *ar, int fudge=0)
    {
        ValVal val = {m_len, 1, fudge, org, ar};
        m_aroffs.push_back(val);
        outb(0);
    }

    void outaroffw(ByteArray *org, ByteArray *ar, int fudge=0)
    {
        ValVal val = {m_len, 2, fudge, org, ar};
        m_aroffs.push_back(val);
        outw(0);
    }

    void outaroffdw(ByteArray *org, ByteArray *ar, int fudge=0)
    {
        ValVal val = {m_len, 4, fudge, org, ar};
        m_aroffs.push_back(val);
        outdw(0);
    }

    void outarlenb(ByteArray *org, ByteArray *ar, int fudge=1)
    {
        ValVal val = {m_len, 1, fudge, org, ar};
        m_arlens.push_back(val);
        outb(0);
    }

    void outarlenw(ByteArray *org, ByteArray *ar, int fudge=1)
    {
        ValVal val = {m_len, 2, fudge, org, ar};
        m_arlens.push_back(val);
        outw(0);
    }

    void outarlendw(ByteArray *org, ByteArray *ar, int fudge=1)
    {
        ValVal val = {m_len, 4, fudge, org, ar};
        m_arlens.push_back(val);
        outdw(0);
    }

    //
    //
    //

    typedef struct {
        WORD pos;
        WORD lenlen;
        WORD fudge;
        ByteArray *origin;
        ByteArray *ar;
    } ValVal;

    void FixupLengths()
    {
        m_fulllen = m_len;

        for ( ArListIt it=m_subars.begin() ; it!=m_subars.end() ; it++ ) {
            (*it)->FixupLengths();
            m_fulllen += (*it)->FullLength();
        }
    }

    WORD GetOffsetOf(ValVal &val)
    {
        if (val.ar==this) return 0;

        WORD clen = m_len;

        for ( ArListIt it=m_subars.begin() ; it!=m_subars.end() ; it++ ) {
            if (val.ar==*it) {
                return clen;
            } else {
                WORD rval = (*it)->GetOffsetOf(val);
                if (rval!=0xffff) {
                    return clen+rval;
                } else {
                    clen += (*it)->FullLength();
                }
            }
        }

        return 0xffff;
    }

    WORD GetLengthOf(ValVal &val)
    {
        if (val.ar==this) return m_fulllen;

        for ( ArListIt it=m_subars.begin() ; it!=m_subars.end() ; it++ ) {
            if (val.ar==*it) {
                return (*it)->FullLength();
            } else {
                WORD t = (*it)->GetLengthOf(val);
                if (t!=0xffff) return t;
            }
        }
        return 0xffff;
    }

    void FixupValues()
    {
        for ( ArListIt ait=m_subars.begin() ; ait!=m_subars.end() ; ait++ ) {
            (*ait)->FixupValues();
        }

        ValListIt vit;

        for ( vit=m_aroffs.begin() ; vit!=m_aroffs.end() ; vit++ ) {
            WORD t = vit->origin->GetOffsetOf(*vit);
            if (t==0xffff) t = 0;
            else t = t + vit->fudge;

            WORD olen = m_len;
            m_len = vit->pos;
            switch (vit->lenlen) {
            case 1:
                outb(t);
                break;
            case 2:
                outw(t);
                break;
            case 4:
                outdw(t);
                break;
            }
            m_len = olen;
        }
        for ( vit=m_arlens.begin() ; vit!=m_arlens.end() ; vit++ ) {
            WORD t = vit->origin->GetLengthOf(*vit);
            if (t==0xffff) t = 0;

            WORD olen = m_len;
            m_len = vit->pos;
            switch (vit->lenlen) {
            case 1:
                outb(t/vit->fudge);
                break;
            case 2:
                outw(t/vit->fudge);
                break;
            case 3:
                outdw(t/vit->fudge);
                break;
            }
            m_len = olen;
        }
    }


    void Fixup()
    {
        FixupLengths();
        FixupValues();
    }

    void Write(HANDLE h)
    {
        DWORD dwWritten = 0;
        WriteFile(h, m_buf, m_len, &dwWritten, NULL);

        for ( ArListIt it=m_subars.begin() ; it!=m_subars.end() ; it++ ) {
            (*it)->Write(h);
        }
    }

protected:
    BYTE *m_buf;
    WORD m_buflen;
    WORD m_len;

    WORD m_fulllen;

    typedef std::vector<ValVal> ValList;
    typedef ValList::iterator ValListIt;

    ValList m_aroffs;
    ValList m_arlens;

    typedef std::vector<ByteArray*> ArList;
    typedef ArList::iterator ArListIt;

    ArList m_subars;
};

//
// United Kingdom standard layout - this has double-quote over 2, pound over 3, and
// an L-shaped Enter key pointing Up and Left. (US has a flat-wide Enter key, many
// other locales have an L-shaped Enter key pointing Left and Down.)
//
// (Sometimes the UK Enter key is also L-shaped pointing Left and Down, in which
// case key OEM_7 is located to the right of OEM_3, instead of to the right of
// OEM_PLUS. It still means the same thing and has the same codes though.)
//
// These are the codes associated with each key, displayed by row in the order each
// symbol appears on the keyboard. VK codes for alphanumeric keys are equal to the
// (uppercase) ASCII value for that symbol, similarly for number keys.
//
// Symbol:             `  1!  2"  3  4$  5%  6^  7&  8*  9(  0)        -_       =+    #~
// Scan code (hex):     29  02  03  04  05  06  07  08  09  0a  0b        0c       0d    2b
// VK code (hex):       df  31  32  33  34  35  36  37  38  39  30        bd       bb    de
// VK name:          OEM_8   1   2   3   4   5   6   7   8   9   0 OEM_MINUS OEM_PLUS OEM_7
//
// Symbol:             Q   W   E   R   T   Y   U   I   O   P    [{    ]}
// Scan code (hex):   10  11  12  13  14  15  16  17  18  19    1a    1b
// VK code (hex):     51  57  45  52  54  59  55  49  4f  50    db    dd
// VK name:            Q   W   E   R   T   Y   U   I   O   P OEM_4 OEM_6
//
// Symbol:             A   S   D   F   G   H   J   K   L    ;:    '@
// Scan code (hex):   1e  1f  20  21  22  23  24  25  26    27    28
// VK code (hex):     41  53  44  46  47  48  4a  4b  4c    ba    c0
// VK name:            A   S   D   F   G   H   J   K   L OEM_1 OEM_3
//
// Symbol:              \|   Z   X   C   V   B   N   M        ,<         .>    /?
// Scan code (hex):     56  2c  2d  2e  2f  30  31  32        33         34    35
// VK code (hex):       dc  5a  58  43  56  42  4e  4d        bc         be    bf
// VK name:          OEM_5   Z   X   C   V   B   N   M OEM_COMMA OEM_PERIOD OEM_2
//
//
// Neither NT nor 9x keyboard layouts concern themselves with the numeric keypad or
// function keys. These are assumed to mean the same across all keyboards.
//

// Scan<->VK<->ANSI translations

const char kbduk_scancodes[] =
    "\x29\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x2b"
    "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b"
    "\x1e\x1f\x20\x21\x22\x23\x24\x25\x26\x27\x28"
    "\x56\x2c\x2d\x2e\x2f\x30\x31\x32\x33\x34\x35";

const char kbduk_vkcodes[] =
    "\xdf" "1234567890\xbd\xbb\xde"
    "QWERTYUIOP\xdb\xdd"
    "ASDFGHJKL\xba\xc0"
    "\xdcZXCVBNM\xbc\xbe\xbf";

const char kbduk_vkansi[] =
    "`1234567890-=#"
    "QWERTYUIOP[]"
    "ASDFGHJKL;'"
    "\\ZXCVBNM,./";

// VK+Shift->codepage translations

const char kbduk_normal_asc[] = // same order as kdbuk_vkcodes
    "`1234567890-=#"
    "qwertyuiop[]"
    "asdfghjkl;'"
    "\\zxcvbnm,./";

const char kbduk_shift_asc[] = // same order as kbduk_vkcodes
    "!\"$%^&*()_+~"
    "QWERTYUIOP{}"
    "ASDFGHJKL:@"
    "|ZXCVBNM<>?";

const char kbduk_ctrl_vk[] =
    "\xdd\xdb\xde\xdc";

const char kbduk_ctrl_asc[] =
    "\x1d\x1b\x1c\x1c";

const char kbduk_ctrlshift_vk[] =
    "6\xbd";

const char kbduk_ctrlshift_asc[] =
    "\x30\x31";

const char kbduk_altgr_vk[] =
    "\xdf" "123456780\xbb\xde"
    "ERTYOP\xdb\xdd"
    "ASD\xba\xc0"
    "\xdcZXCM\xbc\xbe\xbf";

const char kbduk_altgr_asc[] =
    "\xa6\xb9\xb2\xb3\x80\x00^\x00\x00\xba\x00~"
    "\x9c\xae\xfe\xa5\xf8\xb6\x91\x92"
    "\xe6\xdf\xf0\xa8'"
    "`\xf7\xd7\xa9\x99\xb8\xb0\xb4";

const char kbduk_altgrshift_asc[] = // same order as kbduk_altgr_vk
    "\x00\xa1\x84\x00\x80\x89^\x82\xb0\x00\xb1~"
    "\x8c\xae\xde\xa5\xd8\xb6\x93\x94"
    "\xc6\xa7\xd0\xa8\x00"
    "\x00\xf7\xd7\xa9\x99\x8b\x9b\xbf";

// Dead Keys

// (there must also be an entry under the relavent shift table above,
// mapping the VK to the accent char, to match the deadkeys table)
//
// ie, if AltGr+; U gives U-umlaut, then not only must AltGr+; be
// defined as dead key UMLAUT, and UMLAUT+U be defined as U-umlaut,
// both in the tables below, but also AltGr+; must be mapped as the
// real, non-dead key UMLAUT in kbduk_altgr_asc.

#define DK(dkey, vkey, shift) dkey, vkey, shift, 0

#define GRAVE       '`'
#define ACUTE       '\xb4'
#define CIRCUMFLEX  '^'
#define UMLAUT      '\xa8'
#define CEDILLA     '\xb8'
#define RINGABOVE   '\xb0'
#define TILDE       '~'
#define DBLCHAR     '\''

#define SH_NORMAL       0
#define SH_SHIFT        2
#define SH_ALTGR        5

const char kbduk_deadkeys[] = {
    DK(GRAVE, (char)0xdc, SH_ALTGR),
    DK(ACUTE, (char)0xbf, SH_ALTGR),
    DK(CIRCUMFLEX, '6', SH_ALTGR|SH_SHIFT),
    DK(CIRCUMFLEX, '6', SH_ALTGR),
    DK(UMLAUT, (char)0xba, SH_ALTGR|SH_SHIFT),
    DK(UMLAUT, (char)0xba, SH_ALTGR),
    DK(CEDILLA, (char)0xbc, SH_ALTGR),
    DK(RINGABOVE, (char)0xbe, SH_ALTGR),
    DK(TILDE, (char)0xde, SH_ALTGR|SH_SHIFT),
    DK(TILDE, (char)0xde, SH_ALTGR),
    DK(DBLCHAR, (char)0xc0, SH_ALTGR)
};

#define DT(dkey, base, comb) dkey, base, comb

const char kbduk_deadtrans[] = {
    DT(GRAVE, 'a', '\xe0'),
    DT(GRAVE, 'e', '\xe8'),
    DT(GRAVE, 'i', '\xec'),
    DT(GRAVE, 'o', '\xf2'),
    DT(GRAVE, 'u', '\xf9'),
    DT(GRAVE, 'A', '\xc0'),
    DT(GRAVE, 'E', '\xc8'),
    DT(GRAVE, 'I', '\xcc'),
    DT(GRAVE, 'O', '\xd2'),
    DT(GRAVE, 'U', '\xd9'),
    DT(GRAVE, ' ', '\x60'),

    DT(ACUTE, 'a', '\xe1'),
    DT(ACUTE, 'e', '\xe9'),
    DT(ACUTE, 'i', '\xed'),
    DT(ACUTE, 'o', '\xf3'),
    DT(ACUTE, 'u', '\xfa'),
    DT(ACUTE, 'y', '\xfd'),
    DT(ACUTE, 'A', '\xc1'),
    DT(ACUTE, 'E', '\xc9'),
    DT(ACUTE, 'I', '\xcd'),
    DT(ACUTE, 'O', '\xd3'),
    DT(ACUTE, 'U', '\xda'),
    DT(ACUTE, 'Y', '\xdd'),
    DT(ACUTE, ' ', '\xb4'),

    DT(CIRCUMFLEX, 'a', '\xe2'),
    DT(CIRCUMFLEX, 'e', '\xea'),
    DT(CIRCUMFLEX, 'i', '\xee'),
    DT(CIRCUMFLEX, 'o', '\xf4'),
    DT(CIRCUMFLEX, 'u', '\xfb'),
    DT(CIRCUMFLEX, 'A', '\xc2'),
    DT(CIRCUMFLEX, 'E', '\xca'),
    DT(CIRCUMFLEX, 'I', '\xce'),
    DT(CIRCUMFLEX, 'O', '\xd4'),
    DT(CIRCUMFLEX, 'U', '\xdb'),
    DT(CIRCUMFLEX, ' ', '\x5e'),

    DT(UMLAUT, 'a', '\xe4'),
    DT(UMLAUT, 'e', '\xeb'),
    DT(UMLAUT, 'i', '\xef'),
    DT(UMLAUT, 'o', '\xf6'),
    DT(UMLAUT, 'u', '\xfc'),
    DT(UMLAUT, 'y', '\xff'),
    DT(UMLAUT, 'A', '\xc4'),
    DT(UMLAUT, 'E', '\xcb'),
    DT(UMLAUT, 'I', '\xcf'),
    DT(UMLAUT, 'O', '\xd6'),
    DT(UMLAUT, 'U', '\xdc'),
    DT(UMLAUT, 'Y', '\x9f'),
    DT(UMLAUT, ' ', '\xa8'),

    DT(CEDILLA, 'c', '\xe7'),
    DT(CEDILLA, 'C', '\xc7'),
    DT(CEDILLA, ' ', '\xb8'),

    DT(RINGABOVE, 'a', '\xe5'),
    DT(RINGABOVE, 'A', '\xc5'),
    DT(RINGABOVE, ' ', '\xb0'),

    DT(TILDE, 'a', '\xe3'),
    DT(TILDE, 'n', '\xf1'),
    DT(TILDE, 'o', '\xf5'),
    DT(TILDE, 'A', '\xc3'),
    DT(TILDE, 'N', '\xd1'),
    DT(TILDE, 'O', '\xd5'),
    DT(TILDE, ' ', '~'),

    DT(DBLCHAR, '\x8b', '\xab'),
    DT(DBLCHAR, '\x9b', '\xbb')
};

//
// Keyboard layout data structures
// from MSDN\DDK Documentation\Windows 95 Device Driver Kit\Design and Implementation Guide\
//           Windows 95 Keyboard Driver\About the Keyboard Driver\Keyboard Layouts
//


// Header block
typedef struct     {
    BYTE  KbdDesc[2];   // set to 'DS'
    DWORD DefaultLCID;  // base language locale identifier
    WORD  Version;      // set to zero
    DWORD pKbdStart;    // start of the layout data
    WORD  kbdSize;      // number of bytes in the data
    DWORD pDibData;     // must be zero
    WORD  dibSize;      // must be zero
    DWORD pUnicode;      // must be zero
    DWORD nUnicode;      // must be zero
} HeaderBlock;

typedef unsigned short KBDOFFSET; // offset in bytes from start of KbdDataHeader

typedef struct {
    WORD Flags;
    WORD nStateKeys;
    WORD nStates;
    WORD nDeadKeys;
    WORD nLigKeys;
    KBDOFFSET pkbdxxStates;            // STATE_LIST
    KBDOFFSET pkbdxxToAscStates;       //   TOASC_STATES
    KBDOFFSET pkbdxxToAscStateTables;  //   STATETABLES
    KBDOFFSET pkbdxxToAscVkeyList;     // VKEY_LISTS
    KBDOFFSET pkbdxxToAscVKeyListLens; //   VKEY_LIST_LENS
    KBDOFFSET pkbdxxVKShiftStates;     //   VK_STATES
    KBDOFFSET pkbdxxScanToIdx;         // SCANTOIDX
    KBDOFFSET pkbdxxVKeyToIdx;         //   VKEYTOIDX
    WORD      ScanSize;                //   SCAN_SIZE
    KBDOFFSET pkbdxxVKeyToAnsi;        // VKEYTOANSI
    KBDOFFSET pkbdxxDeadKeyTable;      // DEAD_KEYS
    KBDOFFSET pkbdxxDeadTrans;         //   DEAD_KEYTRANS
    KBDOFFSET pkbdxxLigKeys;           // LIG_KEYS KBDOFFSET
    KBDOFFSET pkbdxxCapsBits;          // VKEY CAPSLOCK Table
    KBDOFFSET pkbdxxKanaNormal;        // DBCS platforms (optional)
} KbdDataHeader;

// Guesses from the (lack of) documentation:

//#define ALTGRUSED 1
#define CAPSNORMAL 2
#define SHIFTLOCKUSED 4
//#define DEADCOMBOS 8

// More likely values takes from several .KBD files:
#define ALTGRUSED  0x08
#define DEADCOMBOS 0x20

bool GenKBDFile(LPCTSTR kbdFullPath)
{
    ByteArray s_hdr;
    ByteArray s_khdr;
    ByteArray s_st;
    ByteArray s_tas;
    ByteArray s_tast;
    ByteArray s_tavl;
    ByteArray s_tavll;
    ByteArray s_tats;
    ByteArray s_vkss;
    ByteArray s_s2i;
    ByteArray s_vk2i;
    ByteArray s_vk2a;
    ByteArray s_dkt;
    ByteArray s_dt1;
    ByteArray s_dt2;
    ByteArray s_caps;

    // Construct keytables

    s_st.outb(VK_MENU);
    s_st.outb(0x80);
    s_st.outb(VK_SHIFT);
    s_st.outb(0x80);
    s_st.outb(VK_CONTROL);
    s_st.outb(0x80);

    s_tas.outb(0); // Normal - this is actually the byte before where the KBDOFFSET pointer points
    s_tas.outb(1); // Alt
    s_tas.outb(2); // Shift
    s_tas.outb(3); // Alt+Shift
    s_tas.outb(4); // Ctrl
    s_tas.outb(6); // Ctrl+Shift
    s_tas.outb(5); // AltGr
    s_tas.outb(7); // AltGr+Shift

    // Same order as s_tas, but without the fudged offset
    s_vkss.outb(0);
    s_vkss.outb(0);
    s_vkss.outb(1);
    s_vkss.outb(1);
    s_vkss.outb(2);
    s_vkss.outb(3);
    s_vkss.outb(6);
    s_vkss.outb(7);

    // Scan<->VK<->ANSI translations
    s_s2i.outstr(kbduk_scancodes);
    s_vk2i.outstr(kbduk_vkcodes);
    s_vk2a.outstr(kbduk_vkansi);


    // VK+Shiftstate->codepage translations
    ByteArray s_conv_normal_asc;
    s_conv_normal_asc.outstrn(kbduk_normal_asc, strlen(kbduk_vkcodes));
    ByteArray s_conv_shift_asc;
    s_conv_shift_asc.outstrn(kbduk_shift_asc, strlen(kbduk_vkcodes));

    ByteArray s_conv_ctrl_vk;
    s_conv_ctrl_vk.outstr(kbduk_ctrl_vk);
    ByteArray s_conv_ctrl_asc;
    s_conv_ctrl_asc.outstrn(kbduk_ctrl_asc, strlen(kbduk_ctrl_vk));
    ByteArray s_conv_ctrlshift_vk;
    s_conv_ctrlshift_vk.outstr(kbduk_ctrlshift_vk);
    ByteArray s_conv_ctrlshift_asc;
    s_conv_ctrlshift_asc.outstrn(kbduk_ctrlshift_asc, strlen(kbduk_ctrlshift_vk));

    ByteArray s_conv_altgr_vk;
    s_conv_altgr_vk.outstr(kbduk_altgr_vk);
    ByteArray s_conv_altgr_asc;
    s_conv_altgr_asc.outstrn(kbduk_altgr_asc, strlen(kbduk_altgr_vk));
    ByteArray s_conv_altgrshift_asc;
    s_conv_altgrshift_asc.outstrn(kbduk_altgrshift_asc, strlen(kbduk_altgr_vk));

    s_tats.outar(&s_conv_normal_asc);
    s_tats.outar(&s_conv_shift_asc);
    s_tats.outar(&s_conv_ctrl_vk);
    s_tats.outar(&s_conv_ctrl_asc);
    s_tats.outar(&s_conv_ctrlshift_vk);
    s_tats.outar(&s_conv_ctrlshift_asc);
    s_tats.outar(&s_conv_altgr_vk);
    s_tats.outar(&s_conv_altgr_asc);
    s_tats.outar(&s_conv_altgrshift_asc);

    s_tast.outaroffw(&s_khdr, &s_conv_normal_asc);
    s_tast.outaroffw(&s_khdr, &s_conv_normal_asc);
    s_tast.outaroffw(&s_khdr, &s_conv_shift_asc);
    s_tast.outaroffw(&s_khdr, &s_conv_shift_asc);
    s_tast.outaroffw(&s_khdr, &s_conv_ctrl_asc);
    s_tast.outaroffw(&s_khdr, &s_conv_ctrlshift_asc);
    s_tast.outaroffw(&s_khdr, &s_conv_altgr_asc);
    s_tast.outaroffw(&s_khdr, &s_conv_altgrshift_asc);

    s_tavl.outaroffw(&s_khdr, &s_vk2i);
    s_tavl.outaroffw(&s_khdr, &s_vk2i);
    s_tavl.outaroffw(&s_khdr, &s_vk2i);
    s_tavl.outaroffw(&s_khdr, &s_vk2i);
    s_tavl.outaroffw(&s_khdr, &s_conv_ctrl_vk);
    s_tavl.outaroffw(&s_khdr, &s_conv_ctrlshift_vk);
    s_tavl.outaroffw(&s_khdr, &s_conv_altgr_vk);
    s_tavl.outaroffw(&s_khdr, &s_conv_altgr_vk);

    s_tavll.outarlenb(&s_khdr, &s_vk2i);
    s_tavll.outarlenb(&s_khdr, &s_vk2i);
    s_tavll.outarlenb(&s_khdr, &s_vk2i);
    s_tavll.outarlenb(&s_khdr, &s_vk2i);
    s_tavll.outarlenb(&s_khdr, &s_conv_ctrl_vk);
    s_tavll.outarlenb(&s_khdr, &s_conv_ctrlshift_vk);
    s_tavll.outarlenb(&s_khdr, &s_conv_altgr_vk);
    s_tavll.outarlenb(&s_khdr, &s_conv_altgr_vk);

    // Construct Dead Key tables

    s_dkt.outstrn(kbduk_deadkeys, sizeof(kbduk_deadkeys));

    s_dt1.outw(sizeof(kbduk_deadtrans)/3);
    for ( int i=0 ; i<sizeof(kbduk_deadtrans)/3 ; i++ ) {
        s_dt1.outb(kbduk_deadtrans[3*i]);
        s_dt1.outb(kbduk_deadtrans[3*i+1]);
        s_dt2.outb(kbduk_deadtrans[3*i+2]);
    }

    // Capslock map - Caps key will convert the VKs whose bit is set to uppercase (shifted)
    // all other VKs will remain unshifted
    s_caps.outmb(0, 8);
    s_caps.outb(0xfe);
    s_caps.outb(0xff);
    s_caps.outb(0xff);
    s_caps.outb(0x07);
    s_caps.outmb(0, 20);


    // Build KbdDataHeader stream

    s_khdr.outw(CAPSNORMAL|ALTGRUSED|DEADCOMBOS);
    s_khdr.outw(s_st.Length()/2);
    s_khdr.outw(s_tas.Length()-1);
    s_khdr.outw(s_dkt.Length()/4);
    s_khdr.outw(0);
    s_khdr.outaroffw(&s_khdr, &s_st);
    s_khdr.outaroffw(&s_khdr, &s_tas, 1);
    s_khdr.outaroffw(&s_khdr, &s_tast);
    s_khdr.outaroffw(&s_khdr, &s_tavl);
    s_khdr.outaroffw(&s_khdr, &s_tavll);
    s_khdr.outaroffw(&s_khdr, &s_vkss);
    s_khdr.outaroffw(&s_khdr, &s_s2i);
    s_khdr.outaroffw(&s_khdr, &s_vk2i);
    s_khdr.outarlenw(&s_khdr, &s_s2i);
    s_khdr.outaroffw(&s_khdr, &s_vk2a);
    s_khdr.outaroffw(&s_khdr, &s_dkt);
    s_khdr.outaroffw(&s_khdr, &s_dt1);
    s_khdr.outw(0);
    s_khdr.outaroffw(&s_khdr, &s_caps);

    s_khdr.outar(&s_st);
    s_khdr.outar(&s_tas);
    s_khdr.outar(&s_tast);
    s_khdr.outar(&s_tavl);
    s_khdr.outar(&s_tavll);
    s_khdr.outar(&s_vkss);
    s_khdr.outar(&s_s2i);
    s_khdr.outar(&s_vk2i);
    s_khdr.outar(&s_vk2a);
    s_khdr.outar(&s_tats);
    s_khdr.outar(&s_dkt);
    s_khdr.outar(&s_dt1);
    s_khdr.outar(&s_dt2);
    s_khdr.outar(&s_caps);

    // Build HeaderBlock stream

    s_hdr.outb('D');
    s_hdr.outb('S');
    s_hdr.outdw(kbdLCID);
    s_hdr.outw(0);
    s_hdr.outaroffdw(&s_hdr, &s_khdr);
    s_hdr.outarlenw(&s_hdr, &s_khdr);
    s_hdr.outdw(0);
    s_hdr.outw(0);
    s_hdr.outdw(0);
    s_hdr.outdw(0);
    s_hdr.outar(&s_khdr);

    // Write data to output file

    HANDLE h = CreateFile(kbdFullPath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (h==INVALID_HANDLE_VALUE) return false;

    s_hdr.Fixup();
    s_hdr.Write(h);

    CloseHandle(h);
    return true;
}

bool GenDLLFile(LPCTSTR kbdFullPath)
{
    HRSRC hR = FindResource(NULL, MAKEINTRESOURCE(1), _T("DLLFILE"));
    if (!hR) return false;

    DWORD sz = SizeofResource(NULL, hR);
    if (!sz) return false;

    HGLOBAL hG = LoadResource(NULL, hR);
    if (!hG) return false;

    LPVOID p = LockResource(hG);
    if (!p) return false;

    HANDLE h = CreateFile(kbdFullPath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (h==INVALID_HANDLE_VALUE) return false;

    DWORD dwWritten = 0;
    BOOL rv = WriteFile(h, p, sz, &dwWritten, NULL);

    CloseHandle(h);

    if (!rv || dwWritten!=sz) {
        DeleteFile(kbdFullPath);
        return false;
    }

    return true;
}


bool GenKBD(LPTSTR kbdPath, LPTSTR kbdFullPath, bool bInPlace, bool bNT)
{
    TCHAR sysDir[MAX_PATH];
#ifdef GENTEST
    GetCurrentDirectory(MAX_PATH, sysDir);
#else
    GetSystemDirectory(sysDir, MAX_PATH);
#endif

    CString t1, t2;

    for ( int x=1 ; ; x++ ) {
        if (x<2) {
            t1.Format(_T("%s.%s"), kbdFile, bNT ? _T("DLL") : _T("KBD"));
        } else {
            t1.Format(_T("%s%d.%s"), kbdFile, x, bNT ? _T("DLL") : _T("KBD"));
        }
        t2.Format(_T("%s\\%s"), sysDir, t1);

        DWORD dw = GetFileAttributes(t2);
        if (dw==0xffffffff || bInPlace) {
            if (bNT) {
                if (!GenDLLFile(t2)) return false;
            } else {
                if (!GenKBDFile(t2)) return false;
            }

            _tcscpy(kbdPath, t1);
            _tcscpy(kbdFullPath, t2);
            return true;
        }
    }
    
    return false;
}

int NextLayoutId(CRegKey &klr)
{
    int maxid = 0;

    TCHAR buf[64];
    DWORD i=0, len=0;
    FILETIME ft;
    LONG rv = 0;
    
    while ((rv=RegEnumKeyEx(klr, i++, buf, &(len=64), NULL, NULL, NULL, &ft))==ERROR_SUCCESS) {
        CRegKey kl;
        if (kl.Open(klr, buf, KEY_READ)!=ERROR_SUCCESS) return -1;

        TCHAR buf2[64];
        if (kl.QueryValue(buf2, _T("Layout Id"), &(len=64))==ERROR_SUCCESS) {
            int lid = 0;
            _stscanf(buf2, _T("%x"), &lid);
            if (lid>maxid) maxid = lid;
        }
    }

    return maxid+1;
}

bool RegLayout(LPCTSTR kbdPath, bool bNT)
{
    CRegKey klr;
    if (klr.Open(HKEY_LOCAL_MACHINE, _T("System\\CurrentControlSet\\Control\\Keyboard Layouts"), KEY_ALL_ACCESS)!=ERROR_SUCCESS) return false;

    for ( int lidx=0 ; ; lidx++ ) {
        CString kn;
        kn.Format(_T("%04.4x%04.4x"), lidx, kbdLCID);

        CRegKey kl;
        if (kl.Open(klr, kn)!=ERROR_SUCCESS) {

            int nid = lidx>0 ? NextLayoutId(klr) : 0;
            if (nid==-1) return false;

            if (kl.Create(klr, kn)!=ERROR_SUCCESS) return false;

            kl.SetValue(kbdPath, bNT ? _T("Layout File") : _T("layout file"));
            kl.SetValue(kbdDesc, bNT ? _T("Layout Text") : _T("layout text"));
            if (lidx>0) {
                CString lid;
                lid.Format(_T("%04.4x"), (nid & 0xffff));
                kl.SetValue(lid, bNT ? _T("Layout Id") : _T("layout id"));
            }

            return true;
        }

        // Just keep going until we find the next unused index
    }
}


bool InstallKBD(bool bInPlace, bool bNT)
{
    TCHAR kbdPath[MAX_PATH];
    TCHAR kbdFullPath[MAX_PATH];

    if (!GenKBD(kbdPath, kbdFullPath, bInPlace, bNT)) return false;
#ifndef GENTEST
    if (!bInPlace) {
        if (!RegLayout(kbdPath, bNT)) {
            DeleteFile(kbdFullPath);
            return false;
        }
    }
#else
    ATLTRACE(_T("Created file '%s' at '%s'\n"), kbdPath, kbdFullPath);
#endif

    return true;
}

int KBDInstalled(LPTSTR kbdFullPath, bool bNT)
{
    bool bHaveFile = false;
    bool bHaveReg = false;

    // Check for existing layout file

    TCHAR sysDir[MAX_PATH];
    GetSystemDirectory(sysDir, MAX_PATH);

    CString t1;
    t1.Format(_T("%s\\%s.%s"), sysDir, kbdFile, bNT ? _T("DLL") : _T("KBD"));

    DWORD dw = GetFileAttributes(t1);
    if (dw!=0xffffffff) {
        _tcscpy(kbdFullPath, t1);
        bHaveFile = true;
    }

    // Check for layout registration

    CRegKey klr;
    if (klr.Open(HKEY_LOCAL_MACHINE, _T("System\\CurrentControlSet\\Control\\Keyboard Layouts"), KEY_ALL_ACCESS)==ERROR_SUCCESS) {
        TCHAR buf[64];
        DWORD i=0, len=0;
        FILETIME ft;
        LONG rv = 0;

        while ((rv=RegEnumKeyEx(klr, i++, buf, &(len=64), NULL, NULL, NULL, &ft))==ERROR_SUCCESS) {
            CRegKey kl;
            if (kl.Open(klr, buf, KEY_READ)==ERROR_SUCCESS) {
                TCHAR buf2[64];
                if (kl.QueryValue(buf2, _T("Layout Text"), &(len=64))==ERROR_SUCCESS) {
                    if (_tcsncmp(buf2, kbdDesc, _tcslen(kbdDesc))==0) {
                        bHaveReg = true;
                        break;
                    }
                }
            }
        }
    }

    if (bHaveFile && bHaveReg) return KBD_INSTALLED;
    if (!bHaveFile && !bHaveReg) return KBD_NOTINSTALLED;
    return KBD_PARTIAL;
}

bool GetInstalledKBDs(HWND hLB, bool bNT)
{
    SendMessage(hLB, LB_RESETCONTENT, 0, 0);

    CRegKey klr;
    if (klr.Open(HKEY_LOCAL_MACHINE, _T("System\\CurrentControlSet\\Control\\Keyboard Layouts"), KEY_ALL_ACCESS)==ERROR_SUCCESS) {
        TCHAR buf[64];
        DWORD i=0, len=0;
        FILETIME ft;
        LONG rv = 0;

        while ((rv=RegEnumKeyEx(klr, i++, buf, &(len=64), NULL, NULL, NULL, &ft))==ERROR_SUCCESS) {
            CRegKey kl;
            if (kl.Open(klr, buf, KEY_READ)==ERROR_SUCCESS) {
                TCHAR buf2[64];
                if (kl.QueryValue(buf2, _T("Layout Text"), &(len=64))==ERROR_SUCCESS) {
                    if (_tcsncmp(buf2, kbdDesc, _tcslen(kbdDesc))==0) {
                        TCHAR buf3[MAX_PATH], buf4[MAX_PATH+128];
                        DWORD lid = 0;
                        _stscanf(buf, _T("%x"), &lid);

                        kl.QueryValue(buf3, _T("Layout File"), &(len=MAX_PATH));
                        _sntprintf(buf4, MAX_PATH+128, _T("%s %s (%s)"), buf, buf2, buf3);

                        int idx = SendMessage(hLB, LB_ADDSTRING, 0, (LPARAM)buf4);
                        SendMessage(hLB, LB_SETITEMDATA, idx, lid);
                        SendMessage(hLB, LB_SETSEL, TRUE, idx);
                    }
                }
            }
        }
    }

    return true;
}

bool UninstallKBD(DWORD dwLayoutId)
{
    CRegKey klr;
    if (klr.Open(HKEY_LOCAL_MACHINE, _T("System\\CurrentControlSet\\Control\\Keyboard Layouts"), KEY_ALL_ACCESS)==ERROR_SUCCESS) {

        TCHAR buf[16];
        _sntprintf(buf, 16, _T("%08.8x"), dwLayoutId);

        CRegKey kl;
        if (kl.Open(klr, buf, KEY_READ)==ERROR_SUCCESS) {
            TCHAR buf2[MAX_PATH];
            DWORD i=0, len=0;
            FILETIME ft;
            LONG rv = 0;

            kl.QueryValue(buf2, _T("Layout File"), &(len=MAX_PATH));

            kl.Close();
            klr.DeleteSubKey(buf);

            if (_tcslen(buf2)>0) {
                bool bFileFree = true;

                while ((rv=RegEnumKeyEx(klr, i++, buf, &(len=64), NULL, NULL, NULL, &ft))==ERROR_SUCCESS) {
                    if (kl.Open(klr, buf, KEY_READ)==ERROR_SUCCESS) {
                        TCHAR buf3[MAX_PATH];
                        if (kl.QueryValue(buf3, _T("Layout File"), &(len=64))==ERROR_SUCCESS) {
                            if (_tcsicmp(buf2, buf3)==0) {
                                bFileFree = false;
                                break;
                            }
                        }
                        kl.Close();
                    }
                }

                if (bFileFree) {
                    TCHAR sysDir[MAX_PATH];
                    GetSystemDirectory(sysDir, MAX_PATH);
                    _tcscat(sysDir, _T("\\"));
                    _tcscat(sysDir, buf2);
                    DeleteFile(sysDir);
                }
            }
        }
    }

    return true;
}
